"""
Authentication Vault Resources
* User/Password auth (with DUO)
* AppRole role creation
* AppID
* Policies
* Syslog/File Audit Log
"""
import logging
import yaml
import hvac
import aomi.exceptions
from aomi.vault import wrap_hvac as wrap_vault
from aomi.helpers import hard_path, merge_dicts, cli_hash
from aomi.template import load_var_files, render
from aomi.model.resource import Auth, Resource
from aomi.validation import secret_file, sanitize_mount
LOG = logging.getLogger(__name__)


class DUOAccess(Resource):
    """DUO API
    Access Credentials"""
    child = True

    def export(self, _directory):
        pass

    def secrets(self):
        return [self.secret]

    def obj(self):
        filename = hard_path(self.filename, self.opt.secrets)
        aomi.validation.secret_file(filename)
        obj = yaml.safe_load(open(filename).read())
        return {
            'host': self.host,
            'skey': obj['secret'],
            'ikey': obj['key']
        }

    def diff(self, obj=None):
        return Resource.diff_write_only(self)

    def __init__(self, duo, secret, opt):
        super(DUOAccess, self).__init__({}, opt)
        self.backend = duo.backend
        self.path = "auth/%s/duo/access" % self.backend
        self.filename = secret
        self.secret = secret
        self.host = duo.host

    def fetch(self, vault_client):
        mfa_config = vault_client.read("auth/%s/mfa_config" % self.backend)
        self.existing = mfa_config and mfa_config['data']['type'] == 'duo'


class DUO(Auth):
    """DUO MFA
    Authentication Backend Decorator"""
    required_fields = ['host', 'creds', 'backend']
    resource = 'DUO MFA'
    config_key = 'duo'

    def resources(self):
        return [self, self.access]

    def diff(self, obj=None):
        return Resource.diff_write_only(self)

    def __init__(self, obj, opt):
        super(DUO, self).__init__('userpass', obj, opt)
        self.path = "auth/%s/mfa_config" % self.backend
        self.host = obj['host']
        self.mount = 'userpass'
        self._obj = {'type': 'duo'}
        self.access = DUOAccess(self, obj['creds'], opt)


class AppUser(Resource):
    """App User"""
    required_fields = ['id']
    child = True

    def __init__(self, app, obj, opt):
        super(AppUser, self).__init__(obj, opt)
        self.path = "auth/app-id/map/user-id/%s" % obj['id']
        self._obj = {
            'value': app.app_name
        }
        if 'cidr' in obj:
            self._obj['cidr'] = obj['cidr']


class AppRoleSecret(Resource):
    """Approle Secret"""
    child = True

    def __str__(self):
        return "AppRole Secret %s %s" % (self.role_name, self.secret_name)

    def __init__(self, obj, opt):
        self.role_name = obj['role_name']
        self.secret_name = obj['name']
        self.filename = obj['filename']
        self.opt = opt
        super(AppRoleSecret, self).__init__(obj, opt)

    def diff(self, obj=None):
        return Resource.diff_write_only(self)

    def obj(self):
        filename = hard_path(self.filename, self.opt.secrets)
        aomi.validation.secret_file(filename)
        handle = open(filename, 'r')
        s_obj = {
            'role_name': self.role_name,
            'secret_name': self.secret_name,
            'secret_id': handle.read().strip()
        }
        handle.close()
        return s_obj

    def secrets(self):
        return [self.filename]

    @wrap_vault("writing")
    def write(self, client):
        s_obj = self.obj()
        secret_id = s_obj['secret_id']
        del s_obj['secret_id']
        client.create_role_custom_secret_id(self.role_name,
                                            secret_id,
                                            s_obj)

    @wrap_vault("reading")
    def read(self, client):
        try:
            return client.get_role_secret_id(self.role_name,
                                             self.obj()['secret_id'])
        except hvac.exceptions.InvalidPath:
            return None
        except ValueError as vault_excep:
            if str(vault_excep).startswith('No JSON object'):
                return None

            raise

    @wrap_vault("deleting")
    def delete(self, client):
        client.delete_role_secret_id(self.role_name,
                                     self.obj()['secret_id'])


class AppRole(Auth):
    """AppRole"""
    required_fields = ['name', 'policies']
    config_key = 'approles'

    @staticmethod
    def map_val(src, dest, key, default, src_key=None):
        """Will ensure a dict has values sourced from either
        another dict or based on the provided default"""
        if not src_key:
            src_key = key

        if src_key in src:
            dest[key] = src[src_key]
        else:
            dest[key] = default

    def resources(self):
        return [self] + self.secret_ids

    def __init__(self, obj, opt):
        super(AppRole, self).__init__('approle', obj, opt)
        self.app_name = obj['name']
        self.path = "auth/approle/role/%s" % obj['name']
        self.mount = self.backend
        self.secret_ids = []
        role_obj = {
            'policies': ','.join(obj['policies'])
        }
        AppRole.map_val(role_obj, obj, 'bound_cidr_list', '', 'cidr_list')
        AppRole.map_val(role_obj, obj, 'secret_id_num_uses', 0, 'secret_uses')
        AppRole.map_val(role_obj, obj, 'secret_id_ttl', 0, 'secret_ttl')
        AppRole.map_val(role_obj, obj, 'period', 0)
        AppRole.map_val(role_obj, obj, 'token_max_ttl', 0)
        AppRole.map_val(role_obj, obj, 'token_ttl', 0)
        AppRole.map_val(role_obj, obj, 'bind_secret_id', 'require_secret')
        self._obj = role_obj
        if 'preset' in obj:
            self.presets(obj['preset'], opt)

    def presets(self, presets, opt):
        """Will create representational objects for any preset (push)
        based AppRole Secrets."""
        for preset in presets:
            secret_obj = dict(preset)
            secret_obj['role_name'] = self.app_name
            self.secret_ids.append(AppRoleSecret(secret_obj, opt))

    def diff(self, obj=None):
        obj = dict(self.obj())
        obj['policies'] = obj['policies'].split(',')
        if 'default' not in obj['policies']:
            obj['policies'].append('default')

        obj['policies'] = sorted(obj['policies'])
        return super(AppRole, self).diff(obj)

    @wrap_vault("writing")
    def write(self, client):
        client.create_role(self.app_name, **self.obj())

    @wrap_vault("reading")
    def read(self, client):
        try:
            return client.get_role(self.app_name)
        except hvac.exceptions.InvalidPath:
            return None

    @wrap_vault("deleting")
    def delete(self, client):
        client.delete_role(self.app_name)


class UserPass(Auth):
    """UserPass"""
    required_fields = ['username', 'password_file', 'policies']
    config_key = 'users'

    def export(self, _directory):
        pass

    def __init__(self, obj, opt):
        super(UserPass, self).__init__('userpass', obj, opt)
        self.username = obj['username']
        self.mount = 'userpass'
        self.path = sanitize_mount("auth/userpass/users/%s" % self.username)
        self.policies = obj['policies']
        self.secret = obj['password_file']
        self.filename = self.secret

    def secrets(self):
        return [self.secret]

    def diff(self, obj=None):
        return Resource.diff_write_only(self)

    def obj(self):
        filename = hard_path(self.filename, self.opt.secrets)
        secret_file(filename)
        password = open(filename).readline().strip()
        return {
            'password': password,
            'policies': ','.join(self.policies)
        }


class Policy(Resource):
    """Vault Policy"""
    required_fields = ['file', 'name']
    config_key = 'policies'

    def __init__(self, obj, opt):
        super(Policy, self).__init__(obj, opt)
        self.path = obj['name']
        if self.present:
            self.filename = obj['file']
            cli_obj = merge_dicts(load_var_files(opt),
                                  cli_hash(opt.extra_vars))
            self._obj = merge_dicts(cli_obj, obj.get('vars', {}))

    def validate(self, obj):
        super(Policy, self).validate(obj)
        if 'vars' in obj and not isinstance(obj['vars'], dict):
            raise aomi.exceptions.Validation('policy vars must be dicts')

    def obj(self):
        return render(hard_path(self.filename, self.opt.policies), self._obj) \
            .lstrip() \
            .strip() \
            .replace("\n\n", "\n")

    @wrap_vault("reading")
    def read(self, client):
        LOG.debug("Reading %s", self)
        a_policy = client.get_policy(self.path)
        if a_policy:
            return a_policy.lstrip() \
                           .strip() \
                           .replace("\n\n", "\n")

        return None

    @wrap_vault("writing")
    def write(self, client):
        client.set_policy(self.path, self.obj())

    @wrap_vault("deleting")
    def delete(self, client):
        LOG.debug("Deleting %s", self)
        client.delete_policy(self.path)



""" Vault interactions """
from __future__ import print_function
import os
import socket
import logging
import hvac
import yaml
from aomi.helpers import normalize_vault_path
from aomi.error import output as error_output
from aomi.util import token_file, appid_file
import aomi.error
import aomi.exceptions
LOG = logging.getLogger(__name__)


def approle_token(vault_client, role_id, secret_id):
    """Returns a vault token based on the role and seret id"""
    resp = vault_client.auth_approle(role_id, secret_id)
    if 'auth' in resp and 'client_token' in resp['auth']:
        return resp['auth']['client_token']
    else:
        raise aomi.exceptions.AomiCredentials('invalid approle')


def app_token(vault_client, app_id, user_id):
    """Returns a vault token based on the app and user id."""
    resp = vault_client.auth_app_id(app_id, user_id)
    if 'auth' in resp and 'client_token' in resp['auth']:
        return resp['auth']['client_token']
    else:
        raise aomi.exceptions.AomiCredentials('invalid apptoken')


def token_meta(opt):
    """Generates metadata for a token"""
    meta = {
        'via': 'aomi',
        'operation': opt.operation,
        'hostname': socket.gethostname()
    }
    if 'USER' in os.environ:
        meta['unix_user'] = os.environ['USER']

    if opt.metadata:
        meta_bits = opt.metadata.split(',')
        for meta_bit in meta_bits:
            key, value = meta_bit.split('=')

        if key not in meta:
            meta[key] = value

    for key, value in meta.items():
        LOG.debug("Token metadata %s %s", key, value)

    return meta


def is_mounted(backend, path, backends):
    """Determine whether a backend of a certain type is mounted"""
    for mount_name, values in backends.items():
        b_norm = normalize_vault_path(mount_name)
        m_norm = normalize_vault_path(path)
        if (m_norm == b_norm) and values['type'] == backend:
            return True

    return False


def wrap_hvac(msg):
    """Error catching Vault API wrapper
    This decorator wraps API interactions with Vault. It will
    catch and return appropriate error output on common
    problems"""
    # pylint: disable=missing-docstring
    def wrap_call(func):
        # pylint: disable=missing-docstring
        def func_wrapper(self, vault_client):
            try:
                return func(self, vault_client)
            except (hvac.exceptions.InvalidRequest,
                    hvac.exceptions.Forbidden) as vault_exception:
                if vault_exception.errors[0] == 'permission denied':
                    error_output("Permission denied %s from %s" %
                                 (msg, self.path), self.opt)
                else:
                    raise

        return func_wrapper
    return wrap_call


class Client(hvac.Client):
    """Our Vault Client Wrapper
    This class will pass the existing hvac bits through. When interacting
    with cubbyhole paths, it will use the non operational token in order
    to preserve access."""
    # dat hvac tho
    # pylint: disable=too-many-arguments
    def __init__(self, _url=None, token=None, _cert=None, _verify=True,
                 _timeout=30, _proxies=None, _allow_redirects=True,
                 _session=None):
        if 'VAULT_ADDR' not in os.environ:
            raise aomi.exceptions.AomiError('VAULT_ADDR must be defined')

        ssl_verify = True
        if 'VAULT_SKIP_VERIFY' in os.environ:
            if os.environ['VAULT_SKIP_VERIFY'] == '1':
                ssl_verify = False

        self.initial_token = None
        self.operational_token = None
        super(Client, self).__init__(url=os.environ['VAULT_ADDR'],
                                     verify=ssl_verify)

    def connect(self, opt):
        """This sets up the tokens we expect to see in a way
        that hvac also expects."""
        LOG.info("Connecting to %s", self._url)
        if not self._kwargs['verify']:
            LOG.warning('Skipping SSL Validation!')

        self.token = self.init_token()
        if not self.is_authenticated():
            raise aomi.exceptions.AomiCredentials('initial token')

        if opt.reuse_token:
            LOG.debug("Not creating operational token")
            self.initial_token = self.token
            self.operational_token = self.token
        else:
            self.initial_token = self.token
            self.operational_token = self.op_token(opt)
            if not self.is_authenticated():
                raise aomi.exceptions.AomiCredentials('operational token')

            self.token = self.operational_token

        return self

    def init_token(self):
        """Generate our first token based on workstation configuration"""

        app_filename = appid_file()
        token_filename = token_file()
        if 'VAULT_TOKEN' in os.environ and os.environ['VAULT_TOKEN']:
            LOG.info('Token derived from VAULT_TOKEN environment variable')
            return os.environ['VAULT_TOKEN'].strip()
        elif 'VAULT_USER_ID' in os.environ and \
             'VAULT_APP_ID' in os.environ and \
             os.environ['VAULT_USER_ID'] and os.environ['VAULT_APP_ID']:
            token = app_token(self,
                              os.environ['VAULT_APP_ID'].strip(),
                              os.environ['VAULT_USER_ID'].strip())
            LOG.info("Token derived from VAULT_APP_ID and VAULT_USER_ID")
            return token
        elif 'VAULT_ROLE_ID' in os.environ and \
             'VAULT_SECRET_ID' in os.environ and \
             os.environ['VAULT_ROLE_ID'] and os.environ['VAULT_SECRET_ID']:
            token = approle_token(self,
                                  os.environ['VAULT_ROLE_ID'],
                                  os.environ['VAULT_SECRET_ID'])
            LOG.info("Token derived from VAULT_ROLE_ID and VAULT_SECRET_ID")
            return token
        elif app_filename:
            token = yaml.safe_load(open(app_filename).read().strip())
            if 'app_id' in token and 'user_id' in token:
                token = app_token(self,
                                  token['app_id'],
                                  token['user_id'])
                LOG.info("Token derived from %s", app_filename)
                return token
        elif token_filename:
            LOG.info("Token derived from %s", token_filename)
            return open(token_filename, 'r').read().strip()
        else:
            raise aomi.exceptions.AomiCredentials('unknown method')

    def op_token(self, opt):
        """Return a properly annotated token for our use. This
        token will be revoked at the end of the session. The token
        will have some decent amounts of metadata tho."""
        display_name = self.lookup_token()['data']['display_name']
        args = {
            'lease': opt.lease,
            'display_name': display_name,
            'meta': token_meta(opt)
        }
        try:
            token = self.create_token(**args)
        except (hvac.exceptions.InvalidRequest,
                hvac.exceptions.Forbidden) as vault_exception:
            if vault_exception.errors[0] == 'permission denied':
                error_output("Permission denied creating operational token",
                             opt)
            else:
                raise

        LOG.debug("Using lease of %s", opt.lease)
        return token['auth']['client_token']

    def read(self, path, wrap_ttl=None):
        """Wrap the hvac read call, using the right token for
        cubbyhole interactions."""
        if path.startswith('cubbyhole'):
            self.token = self.initial_token
            val = super(Client, self).read(path, wrap_ttl)
            self.token = self.operational_token
            return val

        return super(Client, self).read(path, wrap_ttl)

    def write(self, path, wrap_ttl=None, **kwargs):
        """Wrap the hvac write call, using the right token for
        cubbyhole interactions."""
        if path.startswith('cubbyhole'):
            self.token = self.initial_token
            val = super(Client, self).write(path, wrap_ttl=wrap_ttl, **kwargs)
            self.token = self.operational_token
            return val
        else:
            super(Client, self).write(path, wrap_ttl=wrap_ttl, **kwargs)

    def delete(self, path):
        """Wrap the hvac delete call, using the right token for
        cubbyhole interactions."""
        if path.startswith('cubbyhole'):
            self.token = self.initial_token
            val = super(Client, self).delete(path)
            self.token = self.operational_token
            return val
        else:
            super(Client, self).delete(path)



"""
Generic Vault Resources
* YAML Variable Files
* Static files
* Generated/Random Secrets
"""
import os
from copy import deepcopy
from uuid import uuid4
import logging
import yaml
from future.utils import iteritems  # pylint: disable=E0401
from cryptorito import portable_b64encode
import aomi.exceptions
from aomi.model.resource import Secret
from aomi.helpers import random_word, hard_path, \
    open_maybe_binary
from aomi.validation import sanitize_mount, secret_file, check_obj, \
    is_unicode_string
LOG = logging.getLogger(__name__)


class Generic(Secret):
    """Generic Secrets"""
    backend = 'generic'

    def __init__(self, obj, opt):
        super(Generic, self).__init__(obj, opt)
        self.mount = sanitize_mount(obj['mount'])
        self.path = "%s/%s" % (self.mount, obj['path'])


class VarFile(Generic):
    """Generic VarFile"""
    required_fields = ['path', 'mount', 'var_file']
    resource_key = 'var_file'

    def secrets(self):
        return [self.secret]

    def __init__(self, obj, opt):
        super(VarFile, self).__init__(obj, opt)
        self.secret = obj['var_file']
        self.filename = obj['var_file']

    def obj(self):
        filename = hard_path(self.filename, self.opt.secrets)
        secret_file(filename)
        return yaml.safe_load(open(filename).read())


class Files(Generic):
    """Generic File"""
    required_fields = ['path', 'mount', 'files']
    resource_key = 'files'

    def secrets(self):
        return [v for _k, v in iteritems(self._obj)]

    def __init__(self, obj, opt):
        super(Files, self).__init__(obj, opt)
        s_obj = {}
        for sfile in obj['files']:
            s_obj[sfile['name']] = sfile['source']

        self._obj = s_obj

    def export(self, directory):
        for name, filename in iteritems(self._obj):
            dest_file = "%s/%s" % (directory, filename)
            dest_dir = os.path.dirname(dest_file)
            if not os.path.isdir(dest_dir):
                os.mkdir(dest_dir, 0o700)

            secret_h = open(dest_file, 'w')
            secret_h.write(self.existing[name])
            secret_h.close()

    def obj(self):
        s_obj = {}
        for name, filename in iteritems(self._obj):
            actual_file = hard_path(filename, self.opt.secrets)
            secret_file(actual_file)
            data = open_maybe_binary(actual_file)
            try:
                is_unicode_string(data)
                s_obj[name] = data
            except aomi.exceptions.Validation:
                s_obj[name] = portable_b64encode(data)
                self.secret_format = 'binary'

        return s_obj

    def validate(self, obj):
        super(Files, self).validate(obj)
        for fileobj in obj['files']:
            check_obj(['source', 'name'], self.name(), fileobj)


def generated_key(key):
    """Create the proper generated key value"""
    key_name = key['name']
    if key['method'] == 'uuid':
        LOG.debug("Setting %s to a uuid", key_name)
        return str(uuid4())
    elif key['method'] == 'words':
        LOG.debug("Setting %s to random words", key_name)
        return random_word()
    elif key['method'] == 'static':
        if 'value' not in key.keys():
            raise aomi.exceptions.AomiData("Missing static value")

        LOG.debug("Setting %s to a static value", key_name)
        return key['value']
    else:
        raise aomi.exceptions.AomiData("Unexpected generated secret method %s"
                                       % key['method'])


class Generated(Generic):
    """Generic Generated"""
    required_fields = ['mount', 'path', 'keys']
    resource_key = 'generated'
    # why are generated generics stored slight differently

    def __init__(self, obj, opt):
        super(Generated, self).__init__(obj['generated'], opt)
        for key in obj['generated']['keys']:
            check_obj(['name', 'method'], 'generated secret entry', key)
        self.keys = obj['generated']['keys']

    def generate_obj(self):
        """Generates the secret object, respecting existing information
        and user specified options"""
        secret_obj = {}
        if self.existing:
            secret_obj = deepcopy(self.existing)

        for key in self.keys:
            key_name = key['name']
            if self.existing and \
               key_name in self.existing and \
               not key.get('overwrite'):
                LOG.debug("Not overwriting %s/%s", self.path, key_name)
                continue
            else:
                secret_obj[key_name] = generated_key(key)

        return secret_obj

    def diff(self, obj=None):
        if self.present and not self.existing:
            return aomi.model.resource.ADD
        elif not self.present and self.existing:
            return aomi.model.resource.DEL
        elif self.present and self.existing:
            overwrites = [x for x in self.keys if x.get('overwrite')]
            if overwrites:
                return aomi.model.resource.OVERWRITE

        return aomi.model.resource.NOOP

    def sync(self, vault_client):
        gen_obj = self.generate_obj()
        self._obj = gen_obj
        super(Generated, self).sync(vault_client)



""" Secret rendering """
from __future__ import print_function
import os
import sys
import logging
from future.utils import iteritems  # pylint: disable=E0401
from pkg_resources import resource_filename
import hvac
from cryptorito import portable_b64decode, is_base64
from aomi.helpers import cli_hash, merge_dicts, \
    path_pieces, abspath
from aomi.template import render, load_var_files
from aomi.error import output as error_output
import aomi.exceptions
LOG = logging.getLogger(__name__)


def grok_seconds(lease):
    """Ensures that we are returning just seconds"""
    if lease.endswith('s'):
        return int(lease[0:-1])
    elif lease.endswith('m'):
        return int(lease[0:-1]) * 60
    elif lease.endswith('h'):
        return int(lease[0:-1]) * 3600

    return None


def is_aws(data):
    """Takes a decent guess as to whether or not we are dealing with
    an AWS secret blob"""
    return 'access_key' in data and 'secret_key' in data


def renew_secret(client, creds, opt):
    """Renews a secret. This will occur unless the user has
    specified on the command line that it is not neccesary"""
    if opt.reuse_token:
        return

    seconds = grok_seconds(opt.lease)
    if not seconds:
        raise aomi.exceptions.AomiCommand("invalid lease %s" % opt.lease)

    renew = client.renew_secret(creds['lease_id'], seconds)
    # sometimes it takes a bit for vault to respond
    # if we are within 5s then we are fine
    if seconds - renew['lease_duration'] >= 5:
        client.revoke_self_token()
        e_msg = 'Unable to renew with desired lease'
        raise aomi.exceptions.VaultConstraint(e_msg)


def secret_key_name(path, key, opt):
    """Renders a Secret key name appropriately"""
    value = key
    if opt.merge_path:
        norm_path = [x for x in path.split('/') if x]
        value = "%s_%s" % ('_'.join(norm_path), key)

    if opt.add_prefix:
        value = "%s%s" % (opt.add_prefix, value)

    if opt.add_suffix:
        value = "%s%s" % (value, opt.add_suffix)

    return value


def grok_template_file(src):
    """Determine the real deal template file"""
    if not src.startswith('builtin:'):
        return abspath(src)

    builtin = src.split(':')[1]
    builtin = "templates/%s.j2" % builtin
    return resource_filename(__name__, builtin)


def blend_vars(secrets, opt):
    """Blends secret and static variables together"""
    extra_obj = merge_dicts(load_var_files(opt),
                            cli_hash(opt.extra_vars))
    merged = merge_dicts(extra_obj, secrets)
    template_obj = dict((k, v) for k, v in iteritems(merged) if v)
    # give templates something to iterate over
    template_obj['aomi_items'] = template_obj.copy()
    return template_obj


def template(client, src, dest, paths, opt):
    """Writes a template using variables from a vault path"""
    key_map = cli_hash(opt.key_map)
    obj = {}
    for path in paths:
        response = client.read(path)
        if not response:
            raise aomi.exceptions.VaultData("Unable to retrieve %s" % path)
        if is_aws(response['data']):
            renew_secret(client, response, opt)

        for s_k, s_v in response['data'].items():
            o_key = s_k
            if s_k in key_map:
                o_key = key_map[s_k]

            k_name = secret_key_name(path, o_key, opt) \
                .lower() \
                .replace('-', '_')
            obj[k_name] = s_v

    template_obj = blend_vars(obj, opt)
    output = render(grok_template_file(src),
                    template_obj)
    write_raw_file(output, abspath(dest))


def write_raw_file(secret, dest):
    """Writes an actual secret out to a file"""
    secret_file = None
    secret_filename = abspath(dest)
    if sys.version_info >= (3, 0):
        if not isinstance(secret, str):
            secret_file = open(secret_filename, 'wb')

    if not secret_file:
        secret_file = open(secret_filename, 'w')

    secret_file.write(secret)
    secret_file.close()
    os.chmod(secret_filename, 0o600)


def raw_file(client, src, dest, opt):
    """Write the contents of a vault path/key to a file. Is
    smart enough to attempt and handle binary files that are
    base64 encoded."""
    path, key = path_pieces(src)
    resp = client.read(path)
    if not resp:
        client.revoke_self_token()
        raise aomi.exceptions.VaultData("Unable to retrieve %s" % path)
    else:
        if 'data' in resp and key in resp['data']:
            secret = resp['data'][key]
            if is_base64(secret):
                LOG.debug('decoding base64 entry')
                secret = portable_b64decode(secret)

            if is_aws(resp['data']):
                renew_secret(client, resp, opt)

            write_raw_file(secret, dest)
        else:
            client.revoke_self_token()
            e_msg = "Key %s not found in %s" % (key, path)
            raise aomi.exceptions.VaultData(e_msg)


def env(client, paths, opt):
    """Renders a shell snippet based on paths in a Secretfile"""
    old_prefix = False
    old_prefix = opt.prefix and not (opt.add_prefix or
                                     opt.add_suffix or
                                     not opt.merge_path)
    if old_prefix:
        LOG.warning("the prefix option is deprecated but being used "
                    "due to not passing in new options")
    elif opt.prefix:
        LOG.warning("the prefix option is deprecated but not being "
                    "used due to passing in new options")
    key_map = cli_hash(opt.key_map)
    for path in paths:
        secrets = client.read(path)
        if secrets and 'data' in secrets:
            if is_aws(secrets['data']):
                renew_secret(client, secrets, opt)

            for s_key, s_val in secrets['data'].items():
                o_key = s_key
                if s_key in key_map:
                    o_key = key_map[s_key]

                # see https://github.com/Autodesk/aomi/issues/40
                env_name = None
                if old_prefix:
                    env_name = ("%s_%s" % (opt.prefix, o_key)).upper()
                else:
                    env_name = secret_key_name(path, o_key, opt).upper()

                print("%s=\"%s\"" % (env_name, s_val))
                if opt.export:
                    print("export %s" % env_name)


def aws(client, path, opt):
    """Renders a shell environment snippet with AWS information"""

    try:
        creds = client.read(path)
    except (hvac.exceptions.InternalServerError) as vault_exception:
        # this is how old vault behaves
        if vault_exception.errors[0].find('unsupported path') > 0:
            error_output("Invalid AWS path. Did you forget the"
                         " credential type and role?", opt)
        else:
            raise

    # this is how new vault behaves
    if not creds:
        error_output("Invalid AWS path. Did you forget the"
                     " credential type and role?", opt)

    renew_secret(client, creds, opt)

    if creds and 'data' in creds:
        print("AWS_ACCESS_KEY_ID=\"%s\"" % creds['data']['access_key'])
        print("AWS_SECRET_ACCESS_KEY=\"%s\"" % creds['data']['secret_key'])
        if 'security_token' in creds['data'] \
           and creds['data']['security_token']:
            token = creds['data']['security_token']
            print("AWS_SECURITY_TOKEN=\"%s\"" % token)
    else:
        client.revoke_self_token()
        e_msg = "Unable to generate AWS credentials from %s" % path
        raise aomi.exceptions.VaultData(e_msg)

    if opt.export:
        print("export AWS_ACCESS_KEY_ID")
        print("export AWS_SECRET_ACCESS_KEY")
        if 'security_token' in creds['data'] \
           and creds['data']['security_token']:
            print("export AWS_SECURITY_TOKEN")



"""Vault Secret Backends"""
import re
import logging
import hvac.exceptions
from aomi.vault import is_mounted
import aomi.exceptions as aomi_excep
from aomi.validation import sanitize_mount
LOG = logging.getLogger(__name__)


class VaultBackend(object):
    """The abstract concept of a Vault backend"""
    backend = None
    list_fun = None
    mount_fun = None
    unmount_fun = None

    @staticmethod
    def list(vault_client, backend_class):
        """List backends"""
        return getattr(vault_client, backend_class.list_fun)()

    def __str__(self):
        if self.backend == self.path:
            return self.backend

        return "%s %s" % (self.backend, self.path)

    def __init__(self, resource, opt):
        self.path = sanitize_mount(resource.mount)
        self.backend = resource.backend
        self.existing = False
        self.present = resource.present
        self.opt = opt

    def sync(self, vault_client):
        """Synchronizes the local and remote Vault resources. Has the net
        effect of adding backend if needed"""
        if not self.existing and self.present:
            self.actually_mount(vault_client)
            LOG.info("Mounting %s backend on %s",
                     self.backend, self.path)
        elif self.existing and self.present:
            LOG.info("%s backend already mounted on %s",
                     self.backend, self.path)
        elif self.existing and not self.present:
            self.unmount(vault_client)
            LOG.info("Unmounting %s backend on %s",
                     self.backend, self.path)
        elif not self.existing and not self.present:
            LOG.info("%s backend already unmounted on %s",
                     self.backend, self.path)

    def fetch(self, backends):
        """Updates local resource with context on whether this
        backend is actually mounted and available"""
        self.existing = is_mounted(self.backend, self.path, backends)

    def unmount(self, client):
        """Unmounts a backend within Vault"""
        getattr(client, self.unmount_fun)(mount_point=self.path)

    def actually_mount(self, client):
        """Actually mount something in Vault"""
        try:
            getattr(client, self.mount_fun)(self.backend,
                                            mount_point=self.path)
        except hvac.exceptions.InvalidRequest as exception:
            match = re.match('existing mount at (?P<path>.+)', str(exception))
            if match:
                e_msg = "%s has a mountpoint conflict with %s" % \
                        (self.path, match.group('path'))
                raise aomi_excep.VaultConstraint(e_msg)
            else:
                raise


class SecretBackend(VaultBackend):
    """Secret Backends for actual Vault resources"""
    list_fun = 'list_secret_backends'
    mount_fun = 'enable_secret_backend'
    unmount_fun = 'disable_secret_backend'


class AuthBackend(VaultBackend):
    """Authentication backends for Vault access"""
    list_fun = 'list_auth_backends'
    mount_fun = 'enable_auth_backend'
    unmount_fun = 'disable_auth_backend'


class LogBackend(VaultBackend):
    """Audit Log backends"""
    list_fun = 'list_audit_backends'
    mount_fun = 'enable_audit_backend'
    unmount_fun = 'disable_audit_backend'

    def __init__(self, resource, opt):
        self.description = None
        super(LogBackend, self).__init__(resource, opt)
        if resource.description:
            self.description = resource.description

        self.obj = resource.obj

    def actually_mount(self, client):
        if self.description:
            client.enable_audit_backend(self.backend,
                                        name=self.backend,
                                        description=self.description,
                                        options=self.obj)
        else:
            client.enable_audit_backend(self.backend,
                                        name=self.backend,
                                        options=self.obj)



"""AWS Secret Backend"""
import logging
import yaml
import aomi.exceptions
import aomi.model.resource
from aomi.vault import is_mounted
from aomi.model.resource import Secret, Resource
from aomi.helpers import hard_path, merge_dicts, cli_hash
from aomi.template import load_var_files, render
from aomi.validation import sanitize_mount, secret_file, check_obj
LOG = logging.getLogger(__name__)


def grok_ttl(secret):
    """Parses the TTL information"""
    ttl_obj = {}
    lease_msg = ''
    if 'lease' in secret:
        ttl_obj['lease'] = secret['lease']
        lease_msg = "lease:%s" % (ttl_obj['lease'])

    if 'lease_max' in secret:
        ttl_obj['lease_max'] = secret['lease_max']
    elif 'lease' in ttl_obj:
        ttl_obj['lease_max'] = ttl_obj['lease']

    if 'lease_max' in ttl_obj:
        lease_msg = "%s lease_max:%s" % (lease_msg, ttl_obj['lease_max'])

    return ttl_obj, lease_msg


class AWSRole(Resource):
    """AWS Role"""
    required_fields = ['name', ['policy', 'arn']]
    child = True

    def __init__(self, mount, obj, opt):
        super(AWSRole, self).__init__(obj, opt)
        self.path = "%s/roles/%s" % (mount, obj['name'])
        if 'policy' in obj:
            self.filename = obj['policy']

        if self.present:
            self._obj = obj
            if 'policy' in self._obj:
                self._obj['policy'] = hard_path(self.filename, opt.policies)

    def export(self, directory):
        if not hasattr(self, 'filename'):
            return

        secret_h = self.export_handle(directory)
        secret_h.write(self.obj()['policy'])
        secret_h.close()

    def obj(self):
        s_obj = {}
        if 'policy' in self._obj:
            role_template_obj = self._obj.get('vars', {})
            cli_obj = merge_dicts(load_var_files(self.opt),
                                  cli_hash(self.opt.extra_vars))
            template_obj = merge_dicts(role_template_obj, cli_obj)
            aws_role = render(self._obj['policy'], template_obj)
            aws_role = aws_role.replace(" ", "").replace("\n", "")
            s_obj = {'policy': aws_role}
        elif 'arn' in self._obj:
            s_obj = {'arn': self._obj['arn']}

        return s_obj


class AWSTTL(Resource):
    """AWS Lease"""
    child = True

    def __init__(self, mount, obj, opt):
        super(AWSTTL, self).__init__(obj, opt)
        self.path = "%s/config/lease" % mount
        self._obj = obj


class AWS(Secret):
    """AWS Backend"""
    resource_key = 'aws_file'
    required_fields = [['aws_file', 'aws'], 'mount',
                       'region', 'roles']
    backend = 'aws'

    def resources(self):
        pieces = [self]
        if self.present:
            pieces = pieces + [self.ttl] + self.roles

        return pieces

    def diff(self, obj=None):
        return Resource.diff_write_only(self)

    def fetch(self, vault_client):
        if is_mounted(self.backend,
                      self.mount,
                      vault_client.list_secret_backends()):
            self.existing = True

    def sync(self, vault_client):
        if self.present:
            LOG.info("Writing AWS root to %s", self.path)
            self.write(vault_client)
        else:
            LOG.info("Removing AWS root at %s", self.path)
            self.delete(vault_client)

    def obj(self):
        _secret, filename, region = self._obj
        actual_filename = hard_path(filename, self.opt.secrets)
        secret_file(actual_filename)
        aws_obj = yaml.safe_load(open(actual_filename, 'r').read())
        check_obj(['access_key_id', 'secret_access_key'],
                  self, aws_obj)
        return {
            'access_key': aws_obj['access_key_id'],
            'secret_key': aws_obj['secret_access_key'],
            'region': region
        }

    def secrets(self):
        return [self._obj[0]]

    def __init__(self, obj, opt):
        super(AWS, self).__init__(obj, opt)
        self.mount = sanitize_mount(obj['mount'])
        self.path = "%s/config/root" % self.mount
        aws_file_path = obj['aws_file']
        if self.present:
            self._obj = (obj['aws_file'],
                         aws_file_path,
                         obj['region'])

            self.roles = []
            for role in obj['roles']:
                self.roles.append(AWSRole(self.mount, role, opt))

            if self.roles is None:
                raise aomi.exceptions.AomiData('missing aws roles')

            ttl_obj, _lease_msg = grok_ttl(obj)
            if ttl_obj:
                self.ttl = AWSTTL(self.mount, ttl_obj, opt)



""" Slightly less terrible error handling and reporting """
from __future__ import print_function
import sys
import traceback


def unhandled(exception, opt):
    """ Handle uncaught/unexpected errors and be polite about it"""
    exmod = type(exception).__module__
    name = "%s.%s" % (exmod, type(exception).__name__)
    # this is a Vault error
    if exmod == 'aomi.exceptions' or exmod == 'cryptorito':
        output(exception.message, opt)
    else:
        output("Unexpected error: %s" % name, opt)


def output(message, opt):
    """ Politely display an unexpected error"""
    print(message, file=sys.stderr)
    if opt.verbose:
        traceback.print_exc(sys.stderr)

    sys.exit(1)



""" The aomi "seed" loop """
from __future__ import print_function
import os
import difflib
from shutil import rmtree
import tempfile
from termcolor import colored
import yaml
from aomi.filez import thaw
from aomi.model import Context
from aomi.template import get_secretfile, render_secretfile
from aomi.model.resource import CHANGED, ADD, DEL, OVERWRITE
from aomi.model.auth import Policy
from aomi.model.aws import AWSRole
from aomi.validation import is_unicode
import aomi.error
import aomi.exceptions


def auto_thaw(opt):
    """Will thaw into a temporary location"""
    icefile = opt.thaw_from
    if not os.path.exists(icefile):
        raise aomi.exceptions.IceFile("%s missing" % icefile)

    thaw(icefile, opt)
    return opt


def seed(vault_client, opt):
    """Will provision vault based on the definition within a Secretfile"""
    if opt.thaw_from:
        opt.secrets = tempfile.mkdtemp('aomi-thaw')
        auto_thaw(opt)

    Context.load(get_secretfile(opt), opt) \
           .fetch(vault_client) \
           .sync(vault_client, opt)

    if opt.thaw_from:
        rmtree(opt.secrets)


def render(directory, opt):
    """Render any provided template. This includes the Secretfile,
    Vault policies, and inline AWS roles"""
    if not os.path.exists(directory) and not os.path.isdir(directory):
        os.mkdir(directory)

    a_secretfile = render_secretfile(opt)
    open("%s/Secretfile" % directory, 'w').write(a_secretfile)
    ctx = Context.load(yaml.safe_load(a_secretfile), opt)
    for resource in ctx.resources():
        if not resource.present:
            continue

        if issubclass(type(resource), Policy):
            if not os.path.isdir("%s/policy" % directory):
                os.mkdir("%s/policy" % directory)
            filename = "%s/policy/%s" % (directory, resource.path)
            open(filename, 'w').write(resource.obj())
        elif issubclass(type(resource), AWSRole):
            if not os.path.isdir("%s/aws" % directory):
                os.mkdir("%s/aws" % directory)
            if 'policy' in resource.obj():
                filename = "%s/aws/%s" % (directory,
                                          os.path.basename(resource.path))
                r_obj = resource.obj()
                if 'policy' in r_obj:
                    open(filename, 'w').write(r_obj['policy'])


def export(vault_client, opt):
    """Export contents of a Secretfile from the Vault server
    into a specified directory."""
    ctx = Context.load(get_secretfile(opt), opt) \
                 .fetch(vault_client)
    for resource in ctx.resources():
        resource.export(opt.directory)


def maybe_colored(msg, color, opt):
    """Maybe it will render in color maybe it will not!"""
    if opt.monochrome:
        return msg

    return colored(msg, color)


def maybe_details(resource, opt):
    """At the first level of verbosity this will print out detailed
    change information on for the specified Vault resource"""
    if opt.verbose == 0:
        return

    if is_unicode(resource.existing):
        a_diff = difflib.unified_diff(resource.existing.splitlines(),
                                      resource.obj().splitlines(),
                                      lineterm="")
        for line in a_diff:
            if line.startswith('+++') or line.startswith('---'):
                continue
            if line[0] == '+':
                print(maybe_colored(line, 'green', opt))
            elif line[0] == '-':
                print(maybe_colored(line, 'red', opt))
            else:
                print(line)


def diff(vault_client, opt):
    """Derive a comparison between what is represented in the Secretfile
    and what is actually live on a Vault instance"""
    if opt.thaw_from:
        opt.secrets = tempfile.mkdtemp('aomi-thaw')
        auto_thaw(opt)
    ctx = Context.load(get_secretfile(opt), opt) \
                 .fetch(vault_client)

    for resource in ctx.resources():
        changed = resource.diff()
        if changed == ADD:
            print("%s %s" % (maybe_colored("+", "green", opt), str(resource)))
        elif changed == DEL:
            print("%s %s" % (maybe_colored("-", "red", opt), str(resource)))
        elif changed == CHANGED:
            print("%s %s" % (maybe_colored("~", "yellow", opt), str(resource)))
        elif changed == OVERWRITE:
            print("%s %s" % (maybe_colored("+", "yellow", opt), str(resource)))

        maybe_details(resource, opt)

    if opt.thaw_from:
        rmtree(opt.secrets)



"""Utilities which are generally that are tied into actual command line
operations processing but do not really fit under seed or render."""
import os
import logging
from aomi.helpers import get_password, path_pieces, \
    backend_type, mount_for_path, abspath
import aomi.validation
import aomi.exceptions
LOG = logging.getLogger(__name__)


def update_user_password(client, userpass):
    """Will update the password for a userpass user"""
    vault_path = ''
    user = ''
    user_path_bits = userpass.split('/')
    if len(user_path_bits) == 1:
        user = user_path_bits[0]
        vault_path = "auth/userpass/users/%s/password" % user
        LOG.debug("Updating password for user %s at the default path", user)
    elif len(user_path_bits) == 2:
        mount = user_path_bits[0]
        user = user_path_bits[1]
        vault_path = "auth/%s/users/%s/password" % (mount, user)
        LOG.debug("Updating password for user %s at path %s", user, mount)
    else:
        client.revoke_self_token()
        raise aomi.exceptions.AomiCommand("invalid user path")

    new_password = get_password()
    obj = {
        'user': user,
        'password': new_password
    }
    client.write(vault_path, **obj)


def update_generic_password(client, path):
    """Will update a single key in a generic secret backend as
    thought it were a password"""
    vault_path, key = path_pieces(path)
    mount = mount_for_path(vault_path, client)
    if not mount:
        client.revoke_self_token()
        raise aomi.exceptions.VaultConstraint('invalid path')

    if backend_type(mount, client) != 'generic':
        client.revoke_self_token()
        raise aomi.exceptions.AomiData("Unsupported backend type")

    LOG.debug("Updating generic password at %s", path)
    existing = client.read(vault_path)
    if not existing or 'data' not in existing:
        LOG.debug("Nothing exists yet at %s!", vault_path)
        existing = {}
    else:
        LOG.debug("Updating %s at %s", key, vault_path)
        existing = existing['data']

    new_password = get_password()
    if key in existing and existing[key] == new_password:
        client.revoke_self_token()
        raise aomi.exceptions.AomiData("Password is same as existing")

    existing[key] = new_password
    client.write(vault_path, **existing)


def password(client, path):
    """Will attempt to contextually update a password in Vault"""
    if path.startswith('user:'):
        update_user_password(client, path[5:])
    else:
        update_generic_password(client, path)


def vault_file(env, default):
    """The path to a misc Vault file
    This function will check for the env override on a file
    path, compute a fully qualified OS appropriate path to
    the desired file and return it if it exists. Otherwise
    returns None
    """
    home = os.environ['HOME'] if 'HOME' in os.environ else \
        os.environ['USERPROFILE']
    filename = os.environ.get(env, os.path.join(home, default))
    filename = abspath(filename)
    if os.path.exists(filename):
        return filename

    return None


def token_file():
    """The path to a Vault Token file"""
    return vault_file('VAULT_TOKEN_FILE', '.vault-token')


def appid_file():
    """The path to an Aomi AppID file"""
    return vault_file('AOMI_APP_FILE', '.aomi-app-token')



""" Helpers for aomi that are used throughout the application """
from __future__ import print_function
import sys
import os
import tempfile
import collections
from random import SystemRandom
from getpass import getpass
import logging
from pkg_resources import resource_string, resource_filename
# Python 2/3 compat
from future.utils import iteritems  # pylint: disable=E0401
import aomi.exceptions
LOG = logging.getLogger()


def my_version():
    """Return the version, checking both packaged and development locations"""
    if os.path.exists(resource_filename(__name__, 'version')):
        return resource_string(__name__, 'version')

    return open(os.path.join(os.path.dirname(__file__),
                             "..", "version")).read()

VERSION = my_version()


def abspath(raw):
    """Return what is hopefully a OS independent path."""
    path_bits = []
    if raw.find('/') != -1:
        path_bits = raw.split('/')
    elif raw.find('\\') != -1:
        path_bits = raw.split('\\')
    else:
        path_bits = [raw]

    return os.path.abspath(os.sep.join(path_bits))


def hard_path(path, prefix_dir):
    """Returns an absolute path to either the relative or absolute file."""
    relative = abspath("%s/%s" % (prefix_dir, path))
    if os.path.exists(relative):
        return relative

    return abspath(path)


def is_tagged(required_tags, has_tags):
    """Checks if tags match"""
    if not required_tags and not has_tags:
        return True
    elif not required_tags:
        return False

    found_tags = []
    for tag in required_tags:
        if tag in has_tags:
            found_tags.append(tag)

    return len(found_tags) == len(required_tags)


def cli_hash(list_of_kv):
    """Parse out a hash from a list of key=value strings"""
    ev_obj = {}
    for extra_var in list_of_kv:
        ev_list = extra_var.split('=')
        key = ev_list[0]
        val = '='.join(ev_list[1:])  # b64 and other side effects
        ev_obj[key] = val

    return ev_obj


def merge_dicts(dict_a, dict_b):
    """Deep merge of two dicts"""
    obj = {}
    for key, value in iteritems(dict_a):
        if key in dict_b:
            if isinstance(dict_b[key], dict):
                obj[key] = merge_dicts(value, dict_b.pop(key))
        else:
            obj[key] = value

    for key, value in iteritems(dict_b):
        obj[key] = value

    return obj


def get_tty_password(confirm):
    """When returning a password from a TTY we assume a user
    is entering it on a keyboard so we ask for confirmation."""
    LOG.debug("Reading password from TTY")
    new_password = getpass('Enter Password: ', stream=sys.stderr)
    if not new_password:
        raise aomi.exceptions.AomiCommand("Must specify a password")

    if not confirm:
        return new_password

    confirm_password = getpass('Again, Please: ', stream=sys.stderr)
    if confirm_password != new_password:
        raise aomi.exceptions.AomiCommand("Passwords do not match")

    return new_password


def get_stdin_password():
    """Returns a password from stdin, no confirmation neccesary"""
    LOG.debug("Reading password from stdin")
    return sys.stdin.readline().rstrip()


def get_password(confirm=True):
    """Will return a password in an appropriate fashion"""
    if sys.stdin.isatty():
        return get_tty_password(confirm)

    return get_stdin_password()


def path_pieces(vault_path):
    """Will return a two part tuple comprising of the vault path
    and the key with in the stored object"""
    path_bits = vault_path.split('/')
    path = '/'.join(path_bits[0:len(path_bits) - 1])
    key = path_bits[len(path_bits) - 1]
    return path, key


def mount_for_path(path, client):
    """Returns the mountpoint for this path"""
    backend_data = client.list_secret_backends()['data']
    backends = [mnt for mnt in backend_data.keys()]
    path_bits = path.split('/')
    if len(path_bits) == 1:
        vault_path = "%s/" % path
        if vault_path in backends:
            return vault_path[0:len(vault_path) - 1]
    else:
        for i in range(1, len(path_bits) + 1):
            vault_path = "%s/" % '/'.join(path_bits[0:i])
            if vault_path in backends:
                return vault_path[0:len(vault_path) - 1]

    return None


def backend_type(path, client):
    """Returns the type of backend at the given mountpoint"""
    backends = client.list_secret_backends()['data']
    vault_path = "%s/" % path
    return backends[vault_path]['type']


def load_word_file(filename):
    """Loads a words file as a list of lines"""
    words_file = resource_filename(__name__, "words/%s" % filename)
    handle = open(words_file, 'r')
    words = handle.readlines()
    handle.close()
    return words


def choose_one(things):
    """Returns a random entry from a list of things"""
    choice = SystemRandom().randint(0, len(things) - 1)
    return things[choice].strip()


def random_word():
    """Returns a random word string"""
    animal = choose_one(load_word_file("animals.txt"))
    academic = choose_one(load_word_file("academic.txt"))
    return "%s-%s" % (academic, animal)


def subdir_path(directory, relative):
    """Returns a file path relative to another path."""
    item_bits = directory.split(os.sep)
    relative_bits = relative.split(os.sep)
    for i in range(0, len(item_bits)):
        if i == len(relative_bits) - 1:
            return os.sep.join(item_bits[i:])
        else:
            if item_bits[i] != relative_bits[i]:
                return None

    return None


def open_maybe_binary(filename):
    """Opens something that might be binary but also
    might be "plain text"."""
    if sys.version_info >= (3, 0):
        data = open(filename, 'rb').read()
        try:
            return data.decode('utf-8')
        except UnicodeDecodeError:
            return data

    return open(filename, 'r').read()


def ensure_dir(path):
    """Ensures a directory exists"""
    if not (os.path.exists(path) and
            os.path.isdir(path)):
        os.mkdir(path)


def ensure_tmpdir():
    """Ensures a temporary directory exists"""
    path = tempfile.mkdtemp('aomi')
    return path


def dict_unicodeize(some_dict):
    """Ensure that every string in a dict is properly represented
    by unicode strings"""

    # some python 2/3 compat
    if isinstance(some_dict, ("".__class__, u"".__class__)):
        if sys.version_info >= (3, 0):
            return some_dict

        return some_dict.decode('utf-8')
    elif isinstance(some_dict, collections.Mapping):
        return dict(map(dict_unicodeize, iteritems(some_dict)))
    elif isinstance(some_dict, collections.Iterable):
        return type(some_dict)(map(dict_unicodeize, some_dict))

    return some_dict


def diff_dict(dict1, dict2):
    """Performs a base type comparison between two dicts"""
    unidict1 = dict_unicodeize(dict1)
    unidict2 = dict_unicodeize(dict2)
    if len(unidict1) != len(unidict2):
        return True

    for comp_k, comp_v in iteritems(unidict1):
        if comp_k not in unidict2:
            return True
        else:
            if comp_v != unidict2[comp_k]:
                return True

    return False


def normalize_vault_path(path):
    """Ensure paths are consistent, always. This covers
    a variety of user specified formats and what HCV
    itself will return in API calls"""
    return '/'.join([x for x in path.split('/') if x])



"""Handle various kinds of import/export of secrets"""
from __future__ import print_function
import os
import shutil
import time
import logging
import datetime
import zipfile
from cryptorito import key_from_keybase, has_gpg_key, \
    import_gpg_key, encrypt, decrypt

from aomi.helpers import subdir_path, ensure_dir, ensure_tmpdir
from aomi.template import get_secretfile
from aomi.validation import gpg_fingerprint \
    as validate_gpg_fingerprint
import aomi.exceptions
from aomi.model import Context
LOG = logging.getLogger(__name__)


def from_keybase(username):
    """Will attempt to retrieve a GPG public key from
    Keybase, importing if neccesary"""
    public_key = key_from_keybase(username)
    fingerprint = public_key['fingerprint'][-8:].upper().encode('ascii')
    key = public_key['bundle'].encode('ascii')
    if not has_gpg_key(fingerprint):
        LOG.debug("Importing gpg key for %s", username)
        if not import_gpg_key(key):
            raise aomi.exceptions.KeybaseAPI("import key for %s" % username)

    return fingerprint


def grok_keys(config):
    """Will retrieve a GPG key from either Keybase or GPG directly"""
    key_ids = []
    for key in config['pgp_keys']:
        if key.startswith('keybase:'):
            key_id = from_keybase(key[8:])
            LOG.debug("Encrypting for keybase user %s", key[8:])
        else:
            if not has_gpg_key(key):
                raise aomi.exceptions.GPG("Do not actually have key %s" % key)

            LOG.debug("Encrypting for gpg id %s", key)
            key_id = key

        validate_gpg_fingerprint(key_id)
        key_ids.append(key_id)

    return key_ids


def freeze_archive(tmp_dir, dest_prefix):
    """Generates a ZIP file of secrets"""
    zip_filename = "%s/aomi-blah.zip" % tmp_dir
    archive = zipfile.ZipFile(zip_filename, 'w')
    for root, _dirnames, filenames in os.walk(dest_prefix):
        for filename in filenames:
            relative_path = subdir_path(root, dest_prefix).split(os.sep)[1:]
            relative_path = os.sep.join(relative_path)
            archive.write("%s/%s" % (root, filename),
                          "%s/%s" % (relative_path, filename))

    archive.close()
    return zip_filename


def freeze_encrypt(dest_dir, zip_filename, config, opt):
    """Encrypts the zip file"""
    pgp_keys = grok_keys(config)
    ice_handle = os.path.basename(os.path.dirname(opt.secretfile))
    timestamp = time.strftime("%H%M%S-%m-%d-%Y",
                              datetime.datetime.now().timetuple())
    ice_file = "%s/aomi-%s-%s.ice" % (dest_dir, ice_handle, timestamp)
    if not encrypt(zip_filename, ice_file, pgp_keys):
        raise aomi.exceptions.GPG("Unable to encrypt zipfile")

    return ice_file


def freeze(dest_dir, opt):
    """Iterates over the Secretfile looking for secrets to freeze"""
    tmp_dir = ensure_tmpdir()
    dest_prefix = "%s/dest" % tmp_dir
    ensure_dir(dest_dir)
    ensure_dir(dest_prefix)
    config = get_secretfile(opt)
    Context.load(config, opt) \
           .freeze(dest_prefix)
    zip_filename = freeze_archive(tmp_dir, dest_prefix)
    ice_file = freeze_encrypt(dest_dir, zip_filename, config, opt)
    shutil.rmtree(tmp_dir)
    LOG.debug("Generated file is %s", ice_file)


def thaw_decrypt(src_file, tmp_dir, opt):
    """Decrypts the encrypted ice file"""

    if not os.path.isdir(opt.secrets):
        LOG.info("Creating secret directory %s", opt.secrets)
        os.mkdir(opt.secrets)

    zip_file = "%s/aomi.zip" % tmp_dir

    if not decrypt(src_file, zip_file):
        raise aomi.exceptions.GPG("Unable to gpg")

    return zip_file


def thaw(src_file, opt):
    """Given the combination of a Secretfile and the output of
    a freeze operation, will restore secrets to usable locations"""
    if not os.path.exists(src_file):
        raise aomi.exceptions.AomiFile("%s does not exist" % src_file)

    tmp_dir = ensure_tmpdir()
    zip_file = thaw_decrypt(src_file, tmp_dir, opt)
    archive = zipfile.ZipFile(zip_file, 'r')
    for archive_file in archive.namelist():
        archive.extract(archive_file, tmp_dir)
        os.chmod("%s/%s" % (tmp_dir, archive_file), 0o640)
        LOG.debug("Extracted %s from archive", archive_file)

    LOG.info("Thawing secrets into %s", opt.secrets)
    config = get_secretfile(opt)
    Context.load(config, opt) \
           .thaw(tmp_dir)



"""Exception definitions for aomi"""


class AomiError(Exception):
    """Our generic exception. Builds up an appropriate error message for
    representation to the user"""
    catmsg = None

    def __init__(self, message=None):
        msg = None
        if self.catmsg is not None and message is not None:
            msg = "%s - %s" % (self.catmsg, message)
        elif self.catmsg is not None:
            msg = self.catmsg
        elif message is not None:
            msg = message

        if msg is not None:
            super(AomiError, self).__init__(msg)
        else:
            super(AomiError, self).__init__()


class AomiCredentials(AomiError):
    """This exception is used for representing errors related to authenticating
    against a running Vault server"""
    catmsg = 'Something wrong with Vault credentials'


class AomiData(AomiError):
    """Some kind of aomi specific data is invalid"""
    catmsg = 'Invalid aomi data'


class AomiCommand(AomiError):
    """Invalid interaction attempted with the aomi cli"""
    catmsg = 'Problem with command line arguments'


class AomiFile(AomiError):
    """Something is wrong with a file on the local filesystem"""
    catmsg = 'Problem with a local file'


class VaultConstraint(AomiError):
    """Vault is imposing constraints on us. Permission or pathing generally"""
    catmsg = 'A Vault Constraint Exists'


class KeybaseAPI(AomiError):
    """Covers errors related to the keybase API"""
    catmsg = 'Something wrong with Keybase integration'


class GPG(AomiError):
    """Covers errors related to our GPG wrapper"""
    catmsg = 'Something went wrong interacting with GPG'


class IceFile(AomiError):
    """Something is wrong with an aomi generated icefile"""
    catmsg = 'Corrupt Icefile'


class VaultData(AomiError):
    """Something is wrong with data received from Vault. Usually
    indicates aomi trying to interact with something manually created"""
    catmsg = 'Unexpected Vault Data Woe'


class Validation(AomiError):
    """Some kind of validation failed. Invalid string, length,
    who knows. Never trust user input tho."""
    catmsg = 'Validation Error'



"""A Model definition of Vault resources"""

#  Ensure every defined model is loaded I guess
import aomi.model.generic
import aomi.model.auth
import aomi.model.aws
import aomi.model.resource
import aomi.model.backend

from aomi.model.context import Context



""" The CLI interface for aomi """
from __future__ import print_function
import os
import sys
import logging
from argparse import ArgumentParser
import aomi.vault
import aomi.render
import aomi.template
import aomi.validation
import aomi.util
import aomi.filez
import aomi.seed_action
from aomi.helpers import VERSION as version
from aomi.util import token_file, appid_file
from aomi.error import unhandled
LOG = logging.getLogger(__name__)


def help_me(parser, opt):
    """Handle display of help and whatever diagnostics"""
    print("aomi v%s" % version)
    print('Get started with aomi'
          ' https://autodesk.github.io/aomi/quickstart')
    if opt.verbose == 2:
        tf_str = 'Token File,' if token_file() else ''
        app_str = 'AppID File,' if appid_file() else ''
        tfe_str = 'Token Env,' if 'VAULT_TOKEN' in os.environ else ''
        appre_str = 'App Role Env,' if 'VAULT_ROLE_ID' in os.environ and \
                    'VAULT_SECRET_ID' in os.environ else ''
        appe_str = 'AppID Env,' if 'VAULT_USER_ID' in os.environ and \
                   'VAULT_APP_ID' in os.environ else ''

        LOG.info(("Auth Hints Present : %s%s%s%s%s" %
                  (tf_str, app_str, tfe_str, appre_str, appe_str))[:-1])
        LOG.info("Vault Server %s" %
                 os.environ['VAULT_ADDR']
                 if 'VAULT_ADDR' in os.environ else '??')

    parser.print_help()
    sys.exit(0)


def extract_file_args(subparsers):
    """Add the command line options for the extract_file operation"""
    extract_parser = subparsers.add_parser('extract_file',
                                           help='Extract a single secret from'
                                           'Vault to a local file')
    extract_parser.add_argument('vault_path',
                                help='Full path (including key) to secret')
    extract_parser.add_argument('destination',
                                help='Location of destination file')
    base_args(extract_parser)


def mapping_args(parser):
    """Add various variable mapping command line options to the parser"""
    parser.add_argument('--add-prefix',
                        dest='add_prefix',
                        help='Specify a prefix to use when '
                        'generating secret key names')
    parser.add_argument('--add-suffix',
                        dest='add_suffix',
                        help='Specify a suffix to use when '
                        'generating secret key names')
    parser.add_argument('--merge-path',
                        dest='merge_path',
                        action='store_true',
                        default=True,
                        help='merge vault path and key name')
    parser.add_argument('--no-merge-path',
                        dest='merge_path',
                        action='store_false',
                        default=True,
                        help='do not merge vault path and key name')
    parser.add_argument('--key-map',
                        dest='key_map',
                        action='append',
                        type=str,
                        default=[])


def export_arg(parser):
    """Add the export argument to a parser"""
    parser.add_argument('--export',
                        dest='export',
                        help='Export declared variables',
                        action='store_true')


def aws_env_args(subparsers):
    """Add command line options for the aws_environment operation"""
    env_parser = subparsers.add_parser('aws_environment')
    env_parser.add_argument('vault_path',
                            help='Full path(s) to the AWS secret')
    export_arg(env_parser)
    base_args(env_parser)


def environment_args(subparsers):
    """Add command line options for the environment operation"""
    env_parser = subparsers.add_parser('environment')
    env_parser.add_argument('vault_paths',
                            help='Full path(s) to secret',
                            nargs='+')
    env_parser.add_argument('--prefix',
                            dest='prefix',
                            help='Old style prefix to use when'
                            'generating secret key names')
    export_arg(env_parser)
    mapping_args(env_parser)
    base_args(env_parser)


def template_args(subparsers):
    """Add command line options for the template operation"""
    template_parser = subparsers.add_parser('template')
    template_parser.add_argument('template',
                                 help='Template source',
                                 nargs='?')
    template_parser.add_argument('destination',
                                 help='Path to write rendered template',
                                 nargs='?')
    template_parser.add_argument('vault_paths',
                                 help='Full path(s) to secret',
                                 nargs='*')
    template_parser.add_argument('--builtin-list',
                                 dest='builtin_list',
                                 help='Display a list of builtin templates',
                                 action='store_true',
                                 default=False)
    template_parser.add_argument('--builtin-info',
                                 dest='builtin_info',
                                 help='Display information on a '
                                 'particular builtin template')
    vars_args(template_parser)
    mapping_args(template_parser)
    base_args(template_parser)


def secretfile_args(parser):
    """Add Secretfile management command line arguments to parser"""
    parser.add_argument('--secrets',
                        dest='secrets',
                        help='Path where secrets are stored',
                        default=os.path.join(os.getcwd(), ".secrets"))
    parser.add_argument('--policies',
                        dest='policies',
                        help='Path where policies are stored',
                        default=os.path.join(os.getcwd(), "vault", ""))
    parser.add_argument('--secretfile',
                        dest='secretfile',
                        help='Secretfile to use',
                        default=os.path.join(os.getcwd(), "Secretfile"))
    parser.add_argument('--tags',
                        dest='tags',
                        help='Tags of things to seed',
                        default=[],
                        type=str,
                        action='append')
    parser.add_argument('--include',
                        dest='include',
                        help='Specify paths to include',
                        default=[],
                        type=str,
                        action='append')
    parser.add_argument('--exclude',
                        dest='exclude',
                        help='Specify paths to exclude',
                        default=[],
                        type=str,
                        action='append')


def generic_args(parser):
    """Command line options associated with every operation
    not just the ones which require connecting to a Vault"""
    parser.add_argument('--verbose',
                        dest='verbose',
                        help='Verbose output',
                        action='count',
                        default=0)


def base_args(parser):
    """Add the generic command line options"""
    generic_args(parser)
    parser.add_argument('--monochrome',
                        dest='monochrome',
                        help='Whether or not to use colors',
                        action='store_true')
    parser.add_argument('--metadata',
                        dest='metadata',
                        help='A series of key=value pairs for token metadata.',
                        default='')
    parser.add_argument('--lease',
                        dest='lease',
                        help='Lease time for intermediary token.',
                        default='10s')
    parser.add_argument('--reuse-token',
                        dest='reuse_token',
                        help='Whether to reuse the existing token. Note'
                        ' this will cause metadata to not be preserved',
                        action='store_true')


def export_args(subparsers):
    """Add command line options for the export operation"""
    export_parser = subparsers.add_parser('export')
    export_parser.add_argument('directory',
                               help='Path where secrets will be exported into')
    secretfile_args(export_parser)
    vars_args(export_parser)
    base_args(export_parser)


def render_args(subparsers):
    """Add command line options for the render operation"""
    render_parser = subparsers.add_parser('render')
    render_parser.add_argument('directory',
                               help='Path where Secrefile and accoutrement'
                               ' will be rendered into')
    secretfile_args(render_parser)
    vars_args(render_parser)
    base_args(render_parser)
    render_parser.add_argument('--thaw-from',
                               dest='thaw_from',
                               help='Thaw an ICE file containing secrets')


def diff_args(subparsers):
    """Add command line options for the diff operation"""
    diff_parser = subparsers.add_parser('diff')
    secretfile_args(diff_parser)
    vars_args(diff_parser)
    base_args(diff_parser)
    diff_parser.add_argument('--thaw-from',
                             dest='thaw_from',
                             help='Thaw an ICE file containing secrets')


def seed_args(subparsers):
    """Add command line options for the seed operation"""
    seed_parser = subparsers.add_parser('seed')
    secretfile_args(seed_parser)
    vars_args(seed_parser)
    seed_parser.add_argument('--mount-only',
                             dest='mount_only',
                             help='Only mount paths if needed',
                             default=False,
                             action='store_true')
    seed_parser.add_argument('--thaw-from',
                             dest='thaw_from',
                             help='Thaw an ICE file containing secrets')
    seed_parser.add_argument('--remove-unknown',
                             dest='remove_unknown',
                             action='store_true',
                             help='Remove mountpoints that are not '
                             'defined in the Secretfile')
    base_args(seed_parser)


def archive_args(parser):
    """Add the command line options for archive related operations"""
    parser.add_argument('icefile',
                        help='Path to the encrypted archive'
                        'file of frozen secrets')


def thaw_args(subparsers):
    """Add command line options for the thaw operation"""
    thaw_parser = subparsers.add_parser('thaw')
    secretfile_args(thaw_parser)
    archive_args(thaw_parser)
    vars_args(thaw_parser)
    base_args(thaw_parser)


def freeze_args(subparsers):
    """Add command line options for the freeze operation"""
    freeze_parser = subparsers.add_parser('freeze')
    secretfile_args(freeze_parser)
    archive_args(freeze_parser)
    vars_args(freeze_parser)
    base_args(freeze_parser)


def password_args(subparsers):
    """Add command line options for the set_password operation"""
    password_parser = subparsers.add_parser('set_password')
    password_parser.add_argument('vault_path',
                                 help='Path which contains password'
                                 'secret to be udpated')
    base_args(password_parser)


def help_args(subparsers):
    """Add command line options for the help operation"""
    help_parser = subparsers.add_parser('help')
    generic_args(help_parser)


def vars_args(parser):
    """Add various command line options for external vars"""
    parser.add_argument('--extra-vars',
                        dest='extra_vars',
                        help='Extra template variables',
                        default=[],
                        type=str,
                        action='append')
    parser.add_argument('--extra-vars-file',
                        dest='extra_vars_file',
                        help='YAML files full of variables',
                        default=[],
                        type=str,
                        action='append')


def token_args(subparsers):
    """Add the CLI options for the token operation"""
    token_parser = subparsers.add_parser('token')
    base_args(token_parser)


def parser_factory(fake_args=None):
    """Return a proper contextual OptionParser"""
    parser = ArgumentParser(description='aomi')
    subparsers = parser.add_subparsers(dest='operation',
                                       help='Specify the data '
                                       ' or extraction operation')
    extract_file_args(subparsers)
    environment_args(subparsers)
    aws_env_args(subparsers)
    seed_args(subparsers)
    render_args(subparsers)
    diff_args(subparsers)
    freeze_args(subparsers)
    thaw_args(subparsers)
    template_args(subparsers)
    password_args(subparsers)
    token_args(subparsers)
    help_args(subparsers)
    export_args(subparsers)

    if fake_args is None:
        return parser, parser.parse_args()

    return parser, parser.parse_args(fake_args)


def template_runner(client, parser, args):
    """Executes template related operations"""
    if args.builtin_list:
        aomi.template.builtin_list()
    elif args.builtin_info:
        aomi.template.builtin_info(args.builtin_info)
    elif args.template and args.destination and args.vault_paths:
        aomi.render.template(client, args.template,
                             args.destination,
                             args.vault_paths,
                             args)
    else:
        parser.print_usage()
        sys.exit(2)

    sys.exit(0)


def ux_actions(parser, args):
    """Handle some human triggers actions"""
    # cryptorito uses native logging (as aomi should tbh)
    normal_fmt = '%(message)s'
    if hasattr(args, 'verbose') and args.verbose and args.verbose >= 2:
        logging.basicConfig(level=logging.DEBUG)
    elif hasattr(args, 'verbose') and args.verbose >= 1:
        logging.basicConfig(level=logging.INFO, format=normal_fmt)
    else:
        logging.basicConfig(level=logging.WARN, format=normal_fmt)

    if args.operation == 'help':
        help_me(parser, args)


def action_runner(parser, args):
    """Run appropriate action, or throw help"""

    ux_actions(parser, args)
    client = aomi.vault.Client(args)
    if args.operation == 'extract_file':
        aomi.render.raw_file(client.connect(args),
                             args.vault_path, args.destination, args)
        sys.exit(0)
    elif args.operation == 'environment':
        aomi.render.env(client.connect(args),
                        args.vault_paths, args)
        sys.exit(0)
    elif args.operation == 'aws_environment':
        aomi.render.aws(client.connect(args),
                        args.vault_path, args)
        sys.exit(0)
    elif args.operation == 'seed':
        aomi.validation.gitignore(args)
        aomi.seed_action.seed(client.connect(args), args)
        sys.exit(0)
    elif args.operation == 'render':
        aomi.seed_action.render(args.directory, args)
        sys.exit(0)
    elif args.operation == 'export':
        aomi.seed_action.export(client.connect(args), args)
        sys.exit(0)
    elif args.operation == 'diff':
        aomi.seed_action.diff(client.connect(args), args)
        sys.exit(0)
    elif args.operation == 'template':
        template_runner(client.connect(args), parser, args)
    elif args.operation == 'token':
        print(client.token)
        sys.exit(0)
    elif args.operation == 'set_password':
        aomi.util.password(client.connect(args), args.vault_path)
        sys.exit(0)
    elif args.operation == 'freeze':
        aomi.filez.freeze(args.icefile, args)
        sys.exit(0)
    elif args.operation == 'thaw':
        aomi.filez.thaw(args.icefile, args)
        sys.exit(0)

    parser.print_usage()
    sys.exit(2)


def main():
    """Entrypoint, sweet Entrypoint"""
    parser, args = parser_factory()
    try:
        action_runner(parser, args)

    # this is our uncaught handler so yes we want to actually
    # catch every error. the format may vary based on the error handler tho
    except Exception as uncaught:  # pylint: disable=broad-except
        unhandled(uncaught, args)
        sys.exit(1)



"""Base Vault Resources"""
import os
import shutil
import logging
import yaml
import hvac.exceptions
from aomi.vault import wrap_hvac as wrap_vault
from aomi.helpers import is_tagged, hard_path, diff_dict
import aomi.exceptions as aomi_excep
from aomi.validation import check_obj, specific_path_check, is_unicode
LOG = logging.getLogger(__name__)

NOOP = 0
CHANGED = 1
ADD = 2
DEL = 3
OVERWRITE = 4


class Resource(object):
    """Vault Resource
    All aomi derived Vault resources should extend this
    class. It provides functionality for validation and
    API CRUD operations."""
    required_fields = []
    config_key = None
    resource_key = None
    child = False
    secret_format = 'data'

    def thaw(self, tmp_dir):
        """Will perform some validation and copy a
        decrypted secret to it's final location"""
        for sfile in self.secrets():
            src_file = "%s/%s" % (tmp_dir, sfile)
            if not os.path.exists(src_file):
                raise aomi_excep \
                    .IceFile("%s secret missing from icefile" %
                             (self))

            dest_file = "%s/%s" % (self.opt.secrets, sfile)
            dest_dir = os.path.dirname(dest_file)
            if not os.path.exists(dest_dir):
                os.mkdir(dest_dir)

            shutil.copy(src_file, dest_file)
            LOG.info("Thawed %s %s", self, sfile)

    def export_handle(self, directory):
        """Get a filehandle for exporting"""
        filename = getattr(self, 'filename')
        dest_file = "%s/%s" % (directory, filename)
        dest_dir = os.path.dirname(dest_file)
        if not os.path.isdir(dest_dir):
            os.mkdir(dest_dir, 0o700)

        return open(dest_file, 'w')

    def export(self, directory):
        """Export exportable resources decoding as needed"""
        if not self.existing or not hasattr(self, 'filename'):
            return

        secret_h = self.export_handle(directory)
        obj = self.existing
        if isinstance(obj, str):
            secret_h.write(obj)
        elif isinstance(obj, dict):
            secret_h.write(yaml.safe_dump(obj))

    def freeze(self, tmp_dir):
        """Copies a secret into a particular location"""
        for sfile in self.secrets():
            src_file = hard_path(sfile, self.opt.secrets)
            if not os.path.exists(src_file):
                raise aomi_excep.IceFile("%s secret not found at %s" %
                                         (self, src_file))

            dest_file = "%s/%s" % (tmp_dir, sfile)
            dest_dir = os.path.dirname(dest_file)
            if not os.path.isdir(dest_dir):
                os.mkdir(dest_dir, 0o700)

            shutil.copy(src_file, dest_file)
            LOG.info("Froze %s %s", self, sfile)

    def resources(self):
        """List of included resources"""
        return [self]

    def grok_state(self, obj):
        """Determine the desired state of this
        resource based on data present"""
        if 'state' in obj:
            my_state = obj['state'].lower()
            if my_state != 'absent' and my_state != 'present':
                raise aomi_excep \
                    .Validation('state must be either "absent" or "present"')

        self.present = obj.get('state', 'present').lower() == 'present'

    def validate(self, obj):
        """Base validation method. Will inspect class attributes
        to dermine just what should be present"""
        if 'tags' in obj and not isinstance(obj['tags'], list):
            raise aomi_excep.Validation('tags must be a list')

        if self.present:
            check_obj(self.required_fields, self.name(), obj)

    def name(self):
        """A Friendly Name for our Resource"""
        return self.__doc__.split('\n')[0]

    def __str__(self):
        return "%s %s" % (self.name(), self.path)

    def obj(self):
        """Returns the Python dict/JSON object representation
        of this Secret as it is to be written to Vault"""
        return self._obj

    # note that this is going to be implemented by subclasses
    def secrets(self):  # pylint: disable=no-self-use
        """Returns a list of secrets which may be used used
        locally by this Vault resource"""
        return []

    def __init__(self, obj, opt):
        self.grok_state(obj)
        self.validate(obj)
        self.path = None
        self.existing = None
        self._obj = {}
        self.tags = obj.get('tags', [])
        self.opt = opt

    def diff(self, obj=None):
        """Determine if something has changed or not"""
        if not self.present:
            if self.existing:
                return DEL

            return NOOP

        if not obj:
            obj = self.obj()

        is_diff = NOOP
        if self.present and self.existing:
            if isinstance(self.existing, dict):
                current = dict(self.existing)
                if 'refresh_interval' in current:
                    del current['refresh_interval']

                if diff_dict(current, obj):
                    is_diff = CHANGED
            elif is_unicode(self.existing):
                if self.existing != obj:
                    is_diff = CHANGED

        elif self.present and not self.existing:
            return ADD

        return is_diff

    def fetch(self, vault_client):
        """Populate internal representation of remote
        Vault resource contents"""
        result = self.read(vault_client)
        if result:
            if isinstance(result, dict) and 'data' in result:
                self.existing = result['data']
            else:
                self.existing = result
        else:
            self.existing = None

    def sync(self, vault_client):
        """Update remove Vault resource contents if needed"""
        if self.present and not self.existing:
            LOG.info("Writing new %s to %s",
                     self.secret_format, self)
            self.write(vault_client)
        elif self.present and self.existing:
            if self.diff() == CHANGED or self.diff() == OVERWRITE:
                LOG.info("Updating %s in %s",
                         self.secret_format, self)
                self.write(vault_client)
        elif not self.present and not self.existing:
            LOG.info("No %s to remove from %s",
                     self.secret_format, self)
        elif not self.present and self.existing:
            LOG.info("Removing %s from %s",
                     self.secret_format, self)
            self.delete(vault_client)

    def filtered(self):
        """Determines whether or not resource is filtered.
        Resources may be filtered if the tags do not match
        or the user has specified explict paths to include
        or exclude via command line options"""
        if not is_tagged(self.tags, self.opt.tags):
            LOG.info("Skipping %s as it does not have requested tags",
                     self.path)
            return False

        if not specific_path_check(self.path, self.opt):
            LOG.info("Skipping %s as it does not match specified paths",
                     self.path)
            return False

        return True

    @staticmethod
    def diff_write_only(resource):
        """A different implementation of diff that is
        used for those Vault resources that are write-only
        such as AWS root configs"""
        if resource.present and not resource.existing:
            return ADD
        elif not resource.present and resource.existing:
            return DEL
        elif resource.present and resource.existing:
            return OVERWRITE

        return NOOP

    @wrap_vault("reading")
    def read(self, client):
        """Read from Vault while handling non surprising errors."""
        LOG.debug("Reading from %s", self)
        try:
            return client.read(self.path)
        except hvac.exceptions.InvalidRequest as vault_exception:
            if str(vault_exception).startswith('no handler for route'):
                return None

    @wrap_vault("writing")
    def write(self, client):
        """Write to Vault while handling non-surprising errors."""
        client.write(self.path, **self.obj())

    @wrap_vault("deleting")
    def delete(self, client):
        """Delete from Vault while handling non-surprising errors."""
        LOG.debug("Deleting %s", self)
        try:
            client.delete(self.path)
        except (hvac.exceptions.InvalidPath,
                hvac.exceptions.InvalidRequest) \
                as vault_exception:
            if str(vault_exception).startswith('no handler for route'):
                return None


class Secret(Resource):
    """Vault Secrets
    These Vault resources will have some kind of secret backend
    underneath them. Seems to work with generic and AWS"""
    config_key = 'secrets'


class Auth(Resource):
    """Auth Backend"""
    def __init__(self, backend, obj, opt):
        super(Auth, self).__init__(obj, opt)
        self.backend = backend


class Mount(Resource):
    """Vault Generic Backend"""
    required_fields = ['path']
    config_key = 'mounts'
    backend = 'generic'
    secret_format = 'mount point'

    def __init__(self, obj, opt):
        super(Mount, self).__init__(obj, opt)
        self.mount = obj['path']
        self.path = self.mount

    def write(self, client):
        return

    def read(self, client):
        return

    def delete(self, client):
        return


class AuditLog(Resource):
    """Audit Logs
    Only supports syslog and file backends"""
    required_fields = ['type']
    config_key = 'audit_logs'

    def write(self, _vault_client):
        pass

    def __init__(self, log_obj, opt):
        super(AuditLog, self).__init__(log_obj, opt)
        self.backend = log_obj['type']
        self.mount = self.backend
        self.description = None
        self.path = log_obj.get('path', self.backend)
        obj = {
            'type': log_obj['type']
        }
        if self.backend == 'file':
            obj['file_path'] = log_obj['file_path']

        if self.backend == 'syslog':
            if 'tag' in log_obj:
                obj['tag'] = log_obj['tag']

            if 'facility' in log_obj:
                obj['facility'] = log_obj['facility']

        if 'description' in obj:
            self.description = obj['description']

        self._obj = obj



"""Some validation helpers for aomi"""
from __future__ import print_function
import sys
import os
import re
import platform
import stat
import logging
from aomi.helpers import abspath, subdir_path
import aomi.exceptions
LOG = logging.getLogger(__name__)


def find_file(name, directory):
    """Searches up from a directory looking for a file"""
    path_bits = directory.split(os.sep)
    for i in range(0, len(path_bits) - 1):
        check_path = path_bits[0:len(path_bits) - i]
        check_file = "%s%s%s" % (os.sep.join(check_path), os.sep, name)
        if os.path.exists(check_file):
            return abspath(check_file)

    return None


def in_file(string, search_file):
    """Looks in a file for a string."""
    handle = open(search_file, 'r')
    for line in handle.readlines():
        if string in line:
            return True

    return False


def gitignore(opt):
    """Will check directories upwards from the Secretfile in order
    to ensure the gitignore file is set properly"""
    directory = os.path.dirname(abspath(opt.secretfile))
    gitignore_file = find_file('.gitignore', directory)
    if gitignore_file:
        secrets_path = subdir_path(abspath(opt.secrets), gitignore_file)
        if secrets_path:
            if not in_file(secrets_path, gitignore_file):
                e_msg = "The path %s was not found in %s" \
                        % (secrets_path, gitignore_file)
                raise aomi.exceptions.AomiFile(e_msg)
        else:
            LOG.debug("Using a non-relative secret directory")

    else:
        raise aomi.exceptions.AomiFile("You should really have a .gitignore")


def secret_file(filename):
    """Will check the permissions of things which really
    should be secret files"""
    filestat = os.stat(abspath(filename))
    if stat.S_ISREG(filestat.st_mode) == 0 and \
       stat.S_ISLNK(filestat.st_mode) == 0:
        e_msg = "Secret file %s must be a real file or symlink" % filename
        raise aomi.exceptions.AomiFile(e_msg)

    if platform.system() != "Windows":
        if filestat.st_mode & stat.S_IROTH or \
           filestat.st_mode & stat.S_IWOTH or \
           filestat.st_mode & stat.S_IWGRP:
            e_msg = "Secret file %s has too loose permissions" % filename
            raise aomi.exceptions.AomiFile(e_msg)


def validate_obj(keys, obj):
    """Super simple "object" validation."""
    msg = ''
    for k in keys:
        if isinstance(k, str):
            if k not in obj or (not isinstance(obj[k], list) and not obj[k]):
                if msg:
                    msg = "%s," % msg

                msg = "%s%s" % (msg, k)
        elif isinstance(k, list):
            found = False
            for k_a in k:
                if k_a in obj:
                    found = True

            if not found:
                if msg:
                    msg = "%s," % msg

                msg = "%s(%s" % (msg, ','.join(k))

    if msg:
        msg = "%s missing" % msg

    return msg


def specific_path_check(path, opt):
    """Will make checks against include/exclude to determine if we
    actually care about the path in question."""
    if opt.exclude:
        if path in opt.exclude:
            return False

    if opt.include:
        if path not in opt.include:
            return False

    return True


def check_obj(keys, name, obj):
    """Do basic validation on an object"""
    msg = validate_obj(keys, obj)

    if msg:
        raise aomi.exceptions.AomiData("object check : %s in %s" % (msg, name))


def sanitize_mount(mount):
    """Returns a quote-unquote sanitized mount path"""
    sanitized_mount = mount
    if sanitized_mount.startswith('/'):
        sanitized_mount = sanitized_mount[1:]

    if sanitized_mount.endswith('/'):
        sanitized_mount = sanitized_mount[:-1]

    return sanitized_mount


def gpg_fingerprint(key):
    """Validates a GPG key fingerprint

    This handles both pre and post GPG 2.1"""
    if (len(key) == 8 and re.match(r'^[0-9A-F]{8}$', key)) or \
       (len(key) == 40 and re.match(r'^[0-9A-F]{40}$', key)):
        return

    raise aomi.exceptions.Validation('Invalid GPG Fingerprint')


def is_unicode_string(string):
    """Validates that we are some kinda unicode string"""
    try:
        if sys.version_info >= (3, 0):
            # isn't a python 3 str actually unicode
            if not isinstance(string, str):
                string.decode('utf-8')

        else:
            string.decode('utf-8')
    except UnicodeError:
        raise aomi.exceptions.Validation('Not a unicode string')


def is_unicode(string):
    """Validates that the object itself is some kinda string"""
    str_type = str(type(string))

    if str_type == 'str' or str_type == 'unicode':
        return True

    return False



#!/usr/bin/env python
import os
import sys
from setuptools import setup
from pkg_resources import resource_filename

# depending on your execution context the version file
# may be located in a different place!
vsn_path = resource_filename(__name__, 'aomi/version')
if not os.path.exists(vsn_path):
    vsn_path = resource_filename(__name__, 'version')
    if not os.path.exists(vsn_path):
        print("%s is missing" % vsn_path)
        sys.exit(1)

setup(name='aomi',
      version=open(vsn_path, 'r').read(),
      description='Easy access to secrets is not as bad as it sounds',
      author='Jonathan Freedman',
      author_email='jonathan.freedman@autodesk.com',
      license='MIT',
      url='https://github.com/autodesk/aomi',
      install_requires=['PyYAML', 'hvac', 'jinja2', 'future',
                        'cryptorito', 'termcolor'],
      include_package_data=True,
      packages=['aomi', 'aomi.model'],
      entry_points={
          'console_scripts': ['aomi = aomi.cli:main']
      },
      package_data={'aomi':['version',
                            'templates/*.j2',
                            'templates/*.yml',
                            'words/*.txt']}
)



""" Render our templates """
from __future__ import print_function
import os
import sys
import traceback
from pkg_resources import resource_listdir, resource_filename
import yaml
from jinja2 import Environment, FileSystemLoader, meta
import jinja2.nodes
import jinja2.exceptions
from cryptorito import portable_b64encode, portable_b64decode
from aomi.helpers import merge_dicts, abspath, cli_hash
import aomi.exceptions as aomi_excep
# Python 2/3 compat
from future.utils import iteritems  # pylint: disable=E0401


def grok_default_vars(parsed_content):
    """Returns a list of vars for which there is a default being set"""
    default_vars = []
    for element in parsed_content.body:
        if isinstance(element, jinja2.nodes.Output):
            for node in element.nodes:
                if isinstance(node, jinja2.nodes.Filter):
                    if node.name == 'default' \
                       and node.node.name not in default_vars:
                        default_vars.append(node.node.name)
        elif isinstance(element, jinja2.nodes.For):
            if isinstance(element.iter, jinja2.nodes.Filter):
                if element.iter.name == 'default' \
                   and element.iter.node.name not in default_vars:
                    default_vars.append(element.iter.node.name)

    return default_vars


def render(filename, obj):
    """Render a template, maybe mixing in extra variables"""
    template_path = abspath(filename)
    fs_loader = FileSystemLoader(os.path.dirname(template_path))
    env = Environment(loader=fs_loader,
                      autoescape=True,
                      trim_blocks=True,
                      lstrip_blocks=True)
    env.filters['b64encode'] = portable_b64encode
    env.filters['b64decode'] = portable_b64decode
    parsed_content = env.parse(env
                               .loader
                               .get_source(env,
                                           os.path.basename(template_path)))

    template_vars = meta.find_undeclared_variables(parsed_content)
    if template_vars:
        missing_vars = []
        default_vars = grok_default_vars(parsed_content)
        for var in template_vars:
            if var not in default_vars and var not in obj:
                missing_vars.append(var)

        if missing_vars:
            e_msg = "Missing required variables %s" % ','.join(missing_vars)
            raise aomi_excep.AomiData(e_msg)

    try:
        return env \
                 .get_template(os.path.basename(template_path)) \
                 .render(**obj)
    except jinja2.exceptions.TemplateSyntaxError as exception:
        template_trace = traceback.format_tb(sys.exc_info()[2])
        raise aomi_excep.Validation("Bad template %s %s" %
                                    (template_trace[len(template_trace) - 1],
                                     str(exception)))
    except jinja2.exceptions.UndefinedError as exception:
        template_traces = [x.strip()
                           for x in traceback.format_tb(sys.exc_info()[2])
                           if 'template code' in x]
        raise aomi_excep.Validation("Missing template variable %s" %
                                    ' '.join(template_traces))


def load_var_files(opt):
    """Load variable files, merge, return contents"""
    obj = {}
    for var_file in opt.extra_vars_file:
        yamlz = yaml.safe_load(open(abspath(var_file)).read())
        obj = merge_dicts(obj.copy(), yamlz)

    return obj


def load_template_help(builtin):
    """Loads the help for a given template"""

    help_file = "templates/%s-help.yml" % builtin
    help_file = resource_filename(__name__, help_file)
    help_obj = {}
    if os.path.exists(help_file):
        help_data = yaml.safe_load(open(help_file))
        if 'name' in help_data:
            help_obj['name'] = help_data['name']

        if 'help' in help_data:
            help_obj['help'] = help_data['help']

        if 'args' in help_data:
            help_obj['args'] = help_data['args']

    return help_obj


def builtin_list():
    """Show a listing of all our builtin templates"""
    for template in resource_listdir(__name__, "templates"):
        builtin, ext = os.path.splitext(os.path.basename(abspath(template)))
        if ext == '.yml':
            continue

        help_obj = load_template_help(builtin)
        if 'name' in help_obj:
            print("%-*s %s" % (20, builtin, help_obj['name']))
        else:
            print("%s" % builtin)


def builtin_info(builtin):
    """Show information on a particular builtin template"""
    help_obj = load_template_help(builtin)
    if help_obj.get('name') and help_obj.get('help'):
        print("The %s template" % (help_obj['name']))
        print(help_obj['help'])
    else:
        print("No help for %s" % builtin)

    if help_obj.get('args'):
        for arg, arg_help in iteritems(help_obj['args']):
            print("  %-*s %s" % (20, arg, arg_help))


def get_secretfile(opt):
    """Returns the de-YAML'd rendered Secretfile"""
    return yaml.safe_load(render_secretfile(opt))


def render_secretfile(opt):
    """Renders and returns the Secretfile construct"""
    secretfile_path = abspath(opt.secretfile)
    obj = merge_dicts(load_var_files(opt),
                      cli_hash(opt.extra_vars))
    return render(secretfile_path, obj)



"""A Context contains a set of Vault resources which may
or may not end up written to the HCV instance. This
context may be filtered, or pre/post processed."""
import sys
import inspect
import logging
from future.utils import iteritems  # pylint: disable=E0401
from aomi.helpers import normalize_vault_path
import aomi.exceptions as aomi_excep
from aomi.model.resource import Resource, Mount, Secret, \
    Auth, AuditLog
from aomi.model.backend import LogBackend, AuthBackend, \
    SecretBackend
LOG = logging.getLogger(__name__)


def filtered_context(context):
    """Filters a context
    This will return a new context with only the resources that
    are actually available for use. Uses tags and command line
    options to make determination."""

    ctx = Context(context.opt)
    for resource in context.resources():
        if resource.child:
            continue
        if resource.filtered():
            ctx.add(resource)

    return ctx


def absent_sort(resource):
    """Used to sort resources in a way where things that
    are being removed are prioritized over things that
    are being added or modified"""
    return resource.present


def find_backend(path, backends):
    """Find the backend at a given path"""
    for backend in backends:
        if backend.path == path:
            return backend

    return None


def ensure_backend(resource, backend, backends, opt):
    """Ensure the backend for a resource is properly in context"""
    existing_mount = find_backend(resource.mount, backends)
    if not existing_mount:
        new_mount = backend(resource, opt)
        backends.append(new_mount)
        return new_mount

    return existing_mount


def find_model(config, obj, mods):
    """Given a list of mods (as returned by py_resources) attempts to
    determine if a given Python obj fits one of the models"""
    for mod in mods:
        if mod[0] != config:
            continue

        if len(mod) == 2:
            return mod[1]

        if len(mod) == 3 and mod[1] in obj:
            return mod[2]

    return None


def py_resources():
    """Discovers all aomi Vault resource models. This includes
    anything extending aomi.model.Mount or aomi.model.Resource."""
    aomi_mods = [m for
                 m, _v in iteritems(sys.modules)
                 if m.startswith('aomi.model')]
    mod_list = []
    mod_map = []
    for amod in [sys.modules[m] for m in aomi_mods]:
        for _mod_bit, model in inspect.getmembers(amod):
            if str(model) in mod_list:
                continue

            if model == Mount:
                mod_list.append(str(model))
                mod_map.append((model.config_key, model))
            elif (inspect.isclass(model) and
                  issubclass(model, Resource) and
                  model.config_key):
                mod_list.append(str(model))
                if model.resource_key:
                    mod_map.append((model.config_key,
                                    model.resource_key,
                                    model))
                elif model.config_key != 'secrets':
                    mod_map.append((model.config_key, model))

    return mod_map


class Context(object):
    """The overall context of an aomi session"""

    @staticmethod
    def load(config, opt):
        """Loads and returns a full context object based on the Secretfile"""
        ctx = Context(opt)
        seed_map = py_resources()
        seed_keys = set([m[0] for m in seed_map])
        for config_key in seed_keys:
            if config_key not in config:
                continue
            for resource in config[config_key]:
                mod = find_model(config_key, resource, seed_map)
                if not mod:
                    LOG.warning("unable to find mod for %s", resource)
                    continue

                ctx.add(mod(resource, opt))

        for config_key in config.keys():
            if config_key != 'pgp_keys' and \
               config_key not in seed_keys:
                LOG.warning("missing model for %s", config_key)

        return filtered_context(ctx)

    def thaw(self, tmp_dir):
        """Will thaw a secret into an appropriate location"""
        for resource in self.resources():
            if resource.present:
                resource.thaw(tmp_dir)

    def freeze(self, dest_dir):
        """Freezes every resource within a context"""
        for resource in self.resources():
            if resource.present:
                resource.freeze(dest_dir)

    def __init__(self, opt):
        self._mounts = []
        self._resources = []
        self._auths = []
        self._logs = []
        self.opt = opt

    def mounts(self):
        """Secret backends within context"""
        return self._mounts

    def logs(self):
        """Audit log backends within context"""
        return self._logs

    def auths(self):
        """Authentication backends within context"""
        return self._auths

    def resources(self):
        """Vault resources within context"""
        res = []
        for resource in self._resources:
            res = res + resource.resources()

        return res

    def add(self, resource):
        """Add a resource to the context"""
        if isinstance(resource, Resource):
            if isinstance(resource, (Secret, Mount)) and \
               resource.mount != 'cubbyhole':
                ensure_backend(resource, SecretBackend, self._mounts, self.opt)
            elif isinstance(resource, (Auth)):
                ensure_backend(resource, AuthBackend, self._auths, self.opt)
            elif isinstance(resource, (AuditLog)):
                ensure_backend(resource, LogBackend, self._logs, self.opt)

            self._resources.append(resource)
        else:
            msg = "Unknown resource %s being " \
                  "added to context" % resource.__class__
            raise aomi_excep.AomiError(msg)

    def remove(self, resource):
        """Removes a resource from the context"""
        if isinstance(resource, Resource):
            self._resources.remove(resource)

    def sync(self, vault_client, opt):
        """Synchronizes the context to the Vault server. This
        has the effect of updating every resource which is
        in the context."""
        active_mounts = []
        for auth in self.auths():
            auth.sync(vault_client)
        for audit_log in self.logs():
            audit_log.sync(vault_client)

        # Handle mounts only on the first pass. This allows us to
        # ensure that everything is in order prior to actually
        # provisioning secrets. Note we handle removals before
        # anything else, allowing us to address mount conflicts.

        # Create a resource set that is only explicit mounts
        # and sort so removals are first
        mounts = [x for x in self.resources()
                  if isinstance(x, (Secret, Mount))]
        s_resources = sorted(mounts, key=absent_sort)
        # Iterate over explicit mounts only
        for resource in s_resources:
            if resource.mount == 'cubbyhole':
                continue

            active_mount = find_backend(resource.mount, active_mounts)
            if not active_mount:
                actual_mount = find_backend(resource.mount, self._mounts)
                active_mounts.append(actual_mount)
                actual_mount.sync(vault_client)

        # Now handle everything else. If "best practices" are being
        # adhered to then every generic mountpoint should exist by now
        not_mounts = [x for x in self.resources()
                      if not isinstance(x, (Mount))]
        for resource in not_mounts:
            resource.sync(vault_client)

        for mount in self.mounts():
            if not find_backend(mount.path, active_mounts):
                mount.unmount(vault_client)

        if opt.remove_unknown:
            self.prune(vault_client)

    def prune(self, vault_client):
        """Will remove any mount point which is not actually defined
        in this context. """
        existing = getattr(vault_client,
                           SecretBackend.list_fun)()['data'].items()
        for mount_name, _values in existing:
            # ignore system paths and cubbyhole
            mount_path = normalize_vault_path(mount_name)
            if mount_path.startswith('sys') or mount_path == 'cubbyhole':
                continue

            exists = [resource.path
                      for resource in self.mounts()
                      if normalize_vault_path(resource.path) == mount_path]

            if not exists:
                LOG.info("removed unknown mount %s", mount_path)
                getattr(vault_client, SecretBackend.unmount_fun)(mount_path)

    def fetch(self, vault_client):
        """Updates the context based on the contents of the Vault
        server. Note that some resources can not be read after
        they have been written to and it is up to those classes
        to handle that case properly."""
        backends = [(self.mounts, SecretBackend),
                    (self.auths, AuthBackend),
                    (self.logs, LogBackend)]
        for b_list, b_class in backends:
            for resource in b_list():
                resource.fetch(getattr(vault_client, b_class.list_fun)())

        for resource in self.resources():
            if issubclass(type(resource), Secret):
                if resource.mount != 'cubbyhole' and \
                   find_backend(resource.mount, self._mounts).existing:
                    resource.fetch(vault_client)
            elif issubclass(type(resource), Auth):
                if find_backend(resource.mount, self._auths).existing:
                    resource.fetch(vault_client)
            elif issubclass(type(resource), Mount):
                resource.existing = find_backend(resource.mount,
                                                 self._mounts).existing
            else:
                resource.fetch(vault_client)

        return self
